# Generated from Pytov.g4 by ANTLR 4.9.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .PytovParser import PytovParser
else:
    from PytovParser import PytovParser

# This class defines a complete listener for a parse tree produced by PytovParser.
class PytovListener(ParseTreeListener):

    # Enter a parse tree produced by PytovParser#parse.
    def enterParse(self, ctx:PytovParser.ParseContext):
        pass

    # Exit a parse tree produced by PytovParser#parse.
    def exitParse(self, ctx:PytovParser.ParseContext):
        pass


    # Enter a parse tree produced by PytovParser#allCode.
    def enterAllCode(self, ctx:PytovParser.AllCodeContext):
        pass

    # Exit a parse tree produced by PytovParser#allCode.
    def exitAllCode(self, ctx:PytovParser.AllCodeContext):
        pass


    # Enter a parse tree produced by PytovParser#seperators.
    def enterSeperators(self, ctx:PytovParser.SeperatorsContext):
        pass

    # Exit a parse tree produced by PytovParser#seperators.
    def exitSeperators(self, ctx:PytovParser.SeperatorsContext):
        pass


    # Enter a parse tree produced by PytovParser#fdivExpression.
    def enterFdivExpression(self, ctx:PytovParser.FdivExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#fdivExpression.
    def exitFdivExpression(self, ctx:PytovParser.FdivExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#powExpression.
    def enterPowExpression(self, ctx:PytovParser.PowExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#powExpression.
    def exitPowExpression(self, ctx:PytovParser.PowExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#modExpression.
    def enterModExpression(self, ctx:PytovParser.ModExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#modExpression.
    def exitModExpression(self, ctx:PytovParser.ModExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#opCpBnExpression.
    def enterOpCpBnExpression(self, ctx:PytovParser.OpCpBnExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#opCpBnExpression.
    def exitOpCpBnExpression(self, ctx:PytovParser.OpCpBnExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#notExpression.
    def enterNotExpression(self, ctx:PytovParser.NotExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#notExpression.
    def exitNotExpression(self, ctx:PytovParser.NotExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#divExpression.
    def enterDivExpression(self, ctx:PytovParser.DivExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#divExpression.
    def exitDivExpression(self, ctx:PytovParser.DivExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#parenExpression.
    def enterParenExpression(self, ctx:PytovParser.ParenExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#parenExpression.
    def exitParenExpression(self, ctx:PytovParser.ParenExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#funcCallExpression.
    def enterFuncCallExpression(self, ctx:PytovParser.FuncCallExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#funcCallExpression.
    def exitFuncCallExpression(self, ctx:PytovParser.FuncCallExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#beforeIncDecExpression.
    def enterBeforeIncDecExpression(self, ctx:PytovParser.BeforeIncDecExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#beforeIncDecExpression.
    def exitBeforeIncDecExpression(self, ctx:PytovParser.BeforeIncDecExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#atomExpression.
    def enterAtomExpression(self, ctx:PytovParser.AtomExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#atomExpression.
    def exitAtomExpression(self, ctx:PytovParser.AtomExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#multExpression.
    def enterMultExpression(self, ctx:PytovParser.MultExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#multExpression.
    def exitMultExpression(self, ctx:PytovParser.MultExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#indexExpression.
    def enterIndexExpression(self, ctx:PytovParser.IndexExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#indexExpression.
    def exitIndexExpression(self, ctx:PytovParser.IndexExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#incDecExpression.
    def enterIncDecExpression(self, ctx:PytovParser.IncDecExpressionContext):
        pass

    # Exit a parse tree produced by PytovParser#incDecExpression.
    def exitIncDecExpression(self, ctx:PytovParser.IncDecExpressionContext):
        pass


    # Enter a parse tree produced by PytovParser#atom.
    def enterAtom(self, ctx:PytovParser.AtomContext):
        pass

    # Exit a parse tree produced by PytovParser#atom.
    def exitAtom(self, ctx:PytovParser.AtomContext):
        pass


    # Enter a parse tree produced by PytovParser#importp.
    def enterImportp(self, ctx:PytovParser.ImportpContext):
        pass

    # Exit a parse tree produced by PytovParser#importp.
    def exitImportp(self, ctx:PytovParser.ImportpContext):
        pass


    # Enter a parse tree produced by PytovParser#listr.
    def enterListr(self, ctx:PytovParser.ListrContext):
        pass

    # Exit a parse tree produced by PytovParser#listr.
    def exitListr(self, ctx:PytovParser.ListrContext):
        pass


    # Enter a parse tree produced by PytovParser#dictKV.
    def enterDictKV(self, ctx:PytovParser.DictKVContext):
        pass

    # Exit a parse tree produced by PytovParser#dictKV.
    def exitDictKV(self, ctx:PytovParser.DictKVContext):
        pass


    # Enter a parse tree produced by PytovParser#dictKVs.
    def enterDictKVs(self, ctx:PytovParser.DictKVsContext):
        pass

    # Exit a parse tree produced by PytovParser#dictKVs.
    def exitDictKVs(self, ctx:PytovParser.DictKVsContext):
        pass


    # Enter a parse tree produced by PytovParser#dictr.
    def enterDictr(self, ctx:PytovParser.DictrContext):
        pass

    # Exit a parse tree produced by PytovParser#dictr.
    def exitDictr(self, ctx:PytovParser.DictrContext):
        pass


    # Enter a parse tree produced by PytovParser#incDec.
    def enterIncDec(self, ctx:PytovParser.IncDecContext):
        pass

    # Exit a parse tree produced by PytovParser#incDec.
    def exitIncDec(self, ctx:PytovParser.IncDecContext):
        pass


    # Enter a parse tree produced by PytovParser#beforeIncDec.
    def enterBeforeIncDec(self, ctx:PytovParser.BeforeIncDecContext):
        pass

    # Exit a parse tree produced by PytovParser#beforeIncDec.
    def exitBeforeIncDec(self, ctx:PytovParser.BeforeIncDecContext):
        pass


    # Enter a parse tree produced by PytovParser#inc.
    def enterInc(self, ctx:PytovParser.IncContext):
        pass

    # Exit a parse tree produced by PytovParser#inc.
    def exitInc(self, ctx:PytovParser.IncContext):
        pass


    # Enter a parse tree produced by PytovParser#dec.
    def enterDec(self, ctx:PytovParser.DecContext):
        pass

    # Exit a parse tree produced by PytovParser#dec.
    def exitDec(self, ctx:PytovParser.DecContext):
        pass


    # Enter a parse tree produced by PytovParser#opCpBn.
    def enterOpCpBn(self, ctx:PytovParser.OpCpBnContext):
        pass

    # Exit a parse tree produced by PytovParser#opCpBn.
    def exitOpCpBn(self, ctx:PytovParser.OpCpBnContext):
        pass


    # Enter a parse tree produced by PytovParser#notl.
    def enterNotl(self, ctx:PytovParser.NotlContext):
        pass

    # Exit a parse tree produced by PytovParser#notl.
    def exitNotl(self, ctx:PytovParser.NotlContext):
        pass


    # Enter a parse tree produced by PytovParser#identifier.
    def enterIdentifier(self, ctx:PytovParser.IdentifierContext):
        pass

    # Exit a parse tree produced by PytovParser#identifier.
    def exitIdentifier(self, ctx:PytovParser.IdentifierContext):
        pass


    # Enter a parse tree produced by PytovParser#decimal.
    def enterDecimal(self, ctx:PytovParser.DecimalContext):
        pass

    # Exit a parse tree produced by PytovParser#decimal.
    def exitDecimal(self, ctx:PytovParser.DecimalContext):
        pass


    # Enter a parse tree produced by PytovParser#string.
    def enterString(self, ctx:PytovParser.StringContext):
        pass

    # Exit a parse tree produced by PytovParser#string.
    def exitString(self, ctx:PytovParser.StringContext):
        pass


    # Enter a parse tree produced by PytovParser#comparator.
    def enterComparator(self, ctx:PytovParser.ComparatorContext):
        pass

    # Exit a parse tree produced by PytovParser#comparator.
    def exitComparator(self, ctx:PytovParser.ComparatorContext):
        pass


    # Enter a parse tree produced by PytovParser#binary.
    def enterBinary(self, ctx:PytovParser.BinaryContext):
        pass

    # Exit a parse tree produced by PytovParser#binary.
    def exitBinary(self, ctx:PytovParser.BinaryContext):
        pass


    # Enter a parse tree produced by PytovParser#operator.
    def enterOperator(self, ctx:PytovParser.OperatorContext):
        pass

    # Exit a parse tree produced by PytovParser#operator.
    def exitOperator(self, ctx:PytovParser.OperatorContext):
        pass


    # Enter a parse tree produced by PytovParser#index.
    def enterIndex(self, ctx:PytovParser.IndexContext):
        pass

    # Exit a parse tree produced by PytovParser#index.
    def exitIndex(self, ctx:PytovParser.IndexContext):
        pass


    # Enter a parse tree produced by PytovParser#indexOverwrite.
    def enterIndexOverwrite(self, ctx:PytovParser.IndexOverwriteContext):
        pass

    # Exit a parse tree produced by PytovParser#indexOverwrite.
    def exitIndexOverwrite(self, ctx:PytovParser.IndexOverwriteContext):
        pass


    # Enter a parse tree produced by PytovParser#boolTF.
    def enterBoolTF(self, ctx:PytovParser.BoolTFContext):
        pass

    # Exit a parse tree produced by PytovParser#boolTF.
    def exitBoolTF(self, ctx:PytovParser.BoolTFContext):
        pass


    # Enter a parse tree produced by PytovParser#ifStatement.
    def enterIfStatement(self, ctx:PytovParser.IfStatementContext):
        pass

    # Exit a parse tree produced by PytovParser#ifStatement.
    def exitIfStatement(self, ctx:PytovParser.IfStatementContext):
        pass


    # Enter a parse tree produced by PytovParser#breakp.
    def enterBreakp(self, ctx:PytovParser.BreakpContext):
        pass

    # Exit a parse tree produced by PytovParser#breakp.
    def exitBreakp(self, ctx:PytovParser.BreakpContext):
        pass


    # Enter a parse tree produced by PytovParser#switchCase.
    def enterSwitchCase(self, ctx:PytovParser.SwitchCaseContext):
        pass

    # Exit a parse tree produced by PytovParser#switchCase.
    def exitSwitchCase(self, ctx:PytovParser.SwitchCaseContext):
        pass


    # Enter a parse tree produced by PytovParser#switchDefault.
    def enterSwitchDefault(self, ctx:PytovParser.SwitchDefaultContext):
        pass

    # Exit a parse tree produced by PytovParser#switchDefault.
    def exitSwitchDefault(self, ctx:PytovParser.SwitchDefaultContext):
        pass


    # Enter a parse tree produced by PytovParser#switchStatement.
    def enterSwitchStatement(self, ctx:PytovParser.SwitchStatementContext):
        pass

    # Exit a parse tree produced by PytovParser#switchStatement.
    def exitSwitchStatement(self, ctx:PytovParser.SwitchStatementContext):
        pass


    # Enter a parse tree produced by PytovParser#identifierList.
    def enterIdentifierList(self, ctx:PytovParser.IdentifierListContext):
        pass

    # Exit a parse tree produced by PytovParser#identifierList.
    def exitIdentifierList(self, ctx:PytovParser.IdentifierListContext):
        pass


    # Enter a parse tree produced by PytovParser#declaration.
    def enterDeclaration(self, ctx:PytovParser.DeclarationContext):
        pass

    # Exit a parse tree produced by PytovParser#declaration.
    def exitDeclaration(self, ctx:PytovParser.DeclarationContext):
        pass


    # Enter a parse tree produced by PytovParser#declarationList.
    def enterDeclarationList(self, ctx:PytovParser.DeclarationListContext):
        pass

    # Exit a parse tree produced by PytovParser#declarationList.
    def exitDeclarationList(self, ctx:PytovParser.DeclarationListContext):
        pass


    # Enter a parse tree produced by PytovParser#variableDeclaration.
    def enterVariableDeclaration(self, ctx:PytovParser.VariableDeclarationContext):
        pass

    # Exit a parse tree produced by PytovParser#variableDeclaration.
    def exitVariableDeclaration(self, ctx:PytovParser.VariableDeclarationContext):
        pass


    # Enter a parse tree produced by PytovParser#expList.
    def enterExpList(self, ctx:PytovParser.ExpListContext):
        pass

    # Exit a parse tree produced by PytovParser#expList.
    def exitExpList(self, ctx:PytovParser.ExpListContext):
        pass


    # Enter a parse tree produced by PytovParser#parameterList.
    def enterParameterList(self, ctx:PytovParser.ParameterListContext):
        pass

    # Exit a parse tree produced by PytovParser#parameterList.
    def exitParameterList(self, ctx:PytovParser.ParameterListContext):
        pass


    # Enter a parse tree produced by PytovParser#parameterInputList.
    def enterParameterInputList(self, ctx:PytovParser.ParameterInputListContext):
        pass

    # Exit a parse tree produced by PytovParser#parameterInputList.
    def exitParameterInputList(self, ctx:PytovParser.ParameterInputListContext):
        pass


    # Enter a parse tree produced by PytovParser#whileStatement.
    def enterWhileStatement(self, ctx:PytovParser.WhileStatementContext):
        pass

    # Exit a parse tree produced by PytovParser#whileStatement.
    def exitWhileStatement(self, ctx:PytovParser.WhileStatementContext):
        pass


    # Enter a parse tree produced by PytovParser#forStatement.
    def enterForStatement(self, ctx:PytovParser.ForStatementContext):
        pass

    # Exit a parse tree produced by PytovParser#forStatement.
    def exitForStatement(self, ctx:PytovParser.ForStatementContext):
        pass


    # Enter a parse tree produced by PytovParser#nonPythonForStatement.
    def enterNonPythonForStatement(self, ctx:PytovParser.NonPythonForStatementContext):
        pass

    # Exit a parse tree produced by PytovParser#nonPythonForStatement.
    def exitNonPythonForStatement(self, ctx:PytovParser.NonPythonForStatementContext):
        pass


    # Enter a parse tree produced by PytovParser#funcDeclaration.
    def enterFuncDeclaration(self, ctx:PytovParser.FuncDeclarationContext):
        pass

    # Exit a parse tree produced by PytovParser#funcDeclaration.
    def exitFuncDeclaration(self, ctx:PytovParser.FuncDeclarationContext):
        pass


    # Enter a parse tree produced by PytovParser#funcCall.
    def enterFuncCall(self, ctx:PytovParser.FuncCallContext):
        pass

    # Exit a parse tree produced by PytovParser#funcCall.
    def exitFuncCall(self, ctx:PytovParser.FuncCallContext):
        pass


    # Enter a parse tree produced by PytovParser#methodCall.
    def enterMethodCall(self, ctx:PytovParser.MethodCallContext):
        pass

    # Exit a parse tree produced by PytovParser#methodCall.
    def exitMethodCall(self, ctx:PytovParser.MethodCallContext):
        pass


    # Enter a parse tree produced by PytovParser#returnp.
    def enterReturnp(self, ctx:PytovParser.ReturnpContext):
        pass

    # Exit a parse tree produced by PytovParser#returnp.
    def exitReturnp(self, ctx:PytovParser.ReturnpContext):
        pass


    # Enter a parse tree produced by PytovParser#statement.
    def enterStatement(self, ctx:PytovParser.StatementContext):
        pass

    # Exit a parse tree produced by PytovParser#statement.
    def exitStatement(self, ctx:PytovParser.StatementContext):
        pass


    # Enter a parse tree produced by PytovParser#block.
    def enterBlock(self, ctx:PytovParser.BlockContext):
        pass

    # Exit a parse tree produced by PytovParser#block.
    def exitBlock(self, ctx:PytovParser.BlockContext):
        pass


    # Enter a parse tree produced by PytovParser#openCurly.
    def enterOpenCurly(self, ctx:PytovParser.OpenCurlyContext):
        pass

    # Exit a parse tree produced by PytovParser#openCurly.
    def exitOpenCurly(self, ctx:PytovParser.OpenCurlyContext):
        pass


    # Enter a parse tree produced by PytovParser#closeCurly.
    def enterCloseCurly(self, ctx:PytovParser.CloseCurlyContext):
        pass

    # Exit a parse tree produced by PytovParser#closeCurly.
    def exitCloseCurly(self, ctx:PytovParser.CloseCurlyContext):
        pass



del PytovParser


#from antlr4 import *
from antlr4.tree.Tree import TerminalNodeImpl
from pytov.PytovVisitor import PytovVisitor
from pytov.PytovParser import PytovParser
from pytov.exceptions import *
from pytov import main
from pytov.Objects import *

class PytovInterpreter(PytovVisitor):
    def __init__(self, file_name, name = 'main'):
        self._name = name
        self.file_name = file_name.replace('\\', '/')

        self.variables = {}
        self.global_variables = {}
        self.errorThrower = LineException(file_name)
        self.builtin_funcs = ['print', 'range', 'len']

        self.block_balance = 0
        self.block_is_true = True
        self.last_block_is_true = True
        self.in_switch = False
        self.in_loop = False
        self.is_in_func = False
        self.active_funcs = []
        self.func_out = None

    def visitTerminal(self, ctx:TerminalNodeImpl):
        if self.is_in_func:
            value = ctx.getText()
            if value in list(self.active_funcs[-1].localVariables.keys()) + list(self.global_variables.keys()):
                return self.active_funcs[-1].localVariables[value]
            return value
        else:
            value = ctx.getText()
            if value in list(self.variables.keys()):
                return self.variables[value]
            elif value in list(self.global_variables.keys()):
                return self.global_variables[value]
            return value

    def visitIndex(self, ctx:PytovParser.IndexContext):
        return self.visit(ctx.children[0])[self.visit(ctx.children[2])]

    def visitIndexOverwrite(self, ctx:PytovParser.IndexOverwriteContext):
        lst = self.visit(ctx.children[0].children[0])
        index = self.visit(ctx.children[0].children[2])
        lst[index] = self.visit(ctx.children[2])

    def visitImportp(self, ctx:PytovParser.ImportpContext):
        if type(ctx.children[1]) == PytovParser.string:
            im_name = ctx.children[1].getText() + ".pv"
        else:
            im_name = self.visit(ctx.children[1]).replace('\\', '/')
        im_name = '/'.join(self.file_name.split('/')[:-1]) + '/' + im_name
        
        imported = main(im_name, 'imported')
        self.variables.update(imported.variables)
        self.global_variables.update(imported.global_variables)

    def visitDeclarationList(self, ctx:PytovParser.DeclarationListContext):
        dec_list = {}
        for dec in ctx.children:
            if type(dec) != TerminalNodeImpl:
                dec_list[dec.children[0].getText()] = self.visit(dec.children[2])
        return dec_list

    def visitIdentifierList(self, ctx:PytovParser.IdentifierListContext):
        id_list = []
        for id in ctx.children:
            if id.getText() != ',':
                id_list.append(id.getText())
        return id_list

    def visitExpList(self, ctx:PytovParser.ExpListContext):
        exp_list = []
        for exp in ctx.children:
            if exp.getText() != ',':
                exp_list.append(self.visit(exp))
        return exp_list

    def visitParameterInputList(self, ctx:PytovParser.ParameterInputListContext):
        params = []
        named = {}
        for i in ctx.children:
            if type(i) != TerminalNodeImpl:
                list_or_dict = self.visit(i)
                if type(i) != PytovParser.ParameterInputListContext:
                    if type(list_or_dict) == list:
                        params.append(list_or_dict)
                    elif type(list_or_dict) == dict:
                        named.update(list_or_dict)
                else:
                    params += list_or_dict[0]
                    named.update(list_or_dict[1])
        if len(params) == 1 and type(params[0]) == list:
            params = params[0]
        return params, named

    def visitParameterList(self, ctx:PytovParser.ParameterListContext):
        params = {}
        for i in ctx.children:
            if type(i) != TerminalNodeImpl:
                list_or_dict = self.visit(i)
                if type(list_or_dict) == list:
                    params.update({i:None for i in list_or_dict})
                elif type(list_or_dict) == dict:
                    params.update(list_or_dict)
        return params

    def getParameterList(self, ctx:PytovParser.ParameterListContext):
        params = []
        named = {}
        for i in ctx.children:
            if type(i) != TerminalNodeImpl:
                if type(i) == PytovParser.ParameterListContext:
                    list_or_dict = self.getParameterList(i)
                    params += list_or_dict[0]
                    named.update(list_or_dict[1])
                else:
                    list_or_dict = self.visit(i)
                    if type(list_or_dict) == list:
                        params += list_or_dict
                    elif type(list_or_dict) == dict:
                        named.update(list_or_dict)
        return params, named

    def visitReturnp(self, ctx:PytovParser.ReturnpContext):
        if self.is_in_func:
            if len(ctx.children) > 1:
                self.func_out = self.visit(ctx.children[1])
            else:
                self.func_out = None
            raise ReturnIndicator("return found.")
        else:
            self.errorThrower._raise([ctx.start.line, ctx.getText()], "SyntaxError", f"You cannot use return outside function.")

    def visitBreakp(self, ctx:PytovParser.BreakpContext):
        if self.in_switch or self.in_loop:
            raise BreakIndicator("break found.")
        else:
            self.errorThrower._raise([ctx.start.line, ctx.getText()], "SyntaxError", f"You cannot use break outside loop or switch.")

    def visitFuncDeclaration(self, ctx:PytovParser.FuncDeclarationContext):
        if len(ctx.children) == 5:
            self.variables[ctx.children[1].getText()] = Func(ctx.children[1].getText(), ctx.children[4], [[],{}], {}) #{'name':ctx.children[1].getText(), 'body':ctx.children[4], 'localVariables':{}, 'args':[[],{}]}
        elif len(ctx.children) == 6:
            self.variables[ctx.children[1].getText()] = Func(ctx.children[1].getText(), ctx.children[5], self.getParameterList(ctx.children[3]), self.visit(ctx.children[3])) #{'name':ctx.children[1].getText(), 'body':ctx.children[5], 'localVariables':self.visit(ctx.children[3]), 'args':self.getParameterList(ctx.children[3])}

    def visitFuncCall(self, ctx:PytovParser.FuncCallContext):
        if ctx.children[0].getText() in list(self.variables.keys()):
            # State outside func
            block_balance = self.block_balance
            block_is_true = self.block_is_true
            last_block_is_true = self.last_block_is_true
            self.block_balance = 0
            self.block_is_true = True
            self.last_block_is_true = True

            func = self.variables[ctx.children[0].getText()]


            func_original_vars = func.localVariables.copy()

            if len(ctx.children) == 4:
                input_args = self.visit(ctx.children[2])
            else:
                input_args = [[], {}]
            
            if len(input_args[0]) != len(func.args[0]):
                print(input_args[0])
                self.errorThrower._raise([ctx.start.line, ctx.getText()], "TypeError", f"'{func.name}' requires {len(func.args[0])} arguments, but {len(input_args[0])} were given.")

            if len(ctx.children) == 4:
                for index, var in enumerate(input_args[0]):
                    func.localVariables[list(func.localVariables.keys())[index]] = var
                func.localVariables.update(input_args[1])


            last_is_in_func = self.is_in_func
            self.is_in_func = True
            self.active_funcs.append(func)
            try:
                self.visit(func.body)
            except ReturnIndicator:
                pass
            self.is_in_func = last_is_in_func
            self.active_funcs.pop(self.active_funcs.index(func))
            func.localVariables = func_original_vars

            # Recreate state
            self.block_balance = block_balance
            self.block_is_true = block_is_true
            self.last_block_is_true = last_block_is_true

            return self.func_out
        elif ctx.children[0].getText() in self.builtin_funcs:
            input_args = self.visit(ctx.children[2])
            
            return eval(ctx.children[0].getText())(*input_args[0], **input_args[1])
        else:
            self.errorThrower._raise([ctx.start.line, ctx.getText()], "NameError", f"Name '{ctx.children[0].getText()}' is not defiend.")

    def visitIfStatement(self, ctx:PytovParser.IfStatementContext):
        ctx.children = [child for child in ctx.children if type(child) != PytovParser.SeperatorsContext]
        if self.visit(ctx.children[1]):
            self.block_is_true = True
            self.visit(ctx.children[2])
            return
        if len(ctx.children) > 3:
            if len(ctx.children) % 3 != 0: # else in if statement
                for i in range(1, (len(ctx.children) - 2) // 3):
                    if self.visit(ctx.children[3*i+1]):
                        self.block_is_true = True
                        self.visit(ctx.children[3*i+2])
                        return
                self.block_is_true = True
                self.visit(ctx.children[-1])
            else:
                for i in range(1, len(ctx.children) // 3):
                    if self.visit(ctx.children[3*i+1]):
                        self.block_is_true = True
                        self.visit(ctx.children[3*i+2])
                        return
    
    def visitSwitchStatement(self, ctx:PytovParser.SwitchStatementContext):
        self.in_switch = True
        ctx.children = [child for child in ctx.children if type(child) != PytovParser.SeperatorsContext]
        switch_on = self.visit(ctx.children[1])
        case_index = 0
        for line in ctx.children:
            if type(line) == PytovParser.SwitchCaseContext:
                if switch_on == self.visit(line.children[1]):
                    try:
                        for run_line in ctx.children[case_index:]:
                            self.visit(run_line)
                    except BreakIndicator:
                        break
            
            elif type(line) == PytovParser.SwitchDefaultContext:
                try:
                    for run_line in ctx.children[case_index:]:
                        self.visit(run_line)
                except BreakIndicator:
                    break
            case_index += 1
        self.in_switch = False

    def visitWhileStatement(self, ctx:PytovParser.WhileStatementContext):
        self.in_loop = True
        try:
            while self.visit(ctx.children[1]):
                self.visit(ctx.children[2])
        except BreakIndicator:
            pass
        self.in_loop = False

    def visitForStatement(self, ctx:PytovParser.ForStatementContext):
        container = self.visit(ctx.children[3])
        self.in_loop = True
        try:
            for i in range(len(container)):
                if self.is_in_func:
                    self.active_funcs[-1].localVariables[ctx.children[1].getText()] = container[i]
                else:
                    self.variables[ctx.children[1].getText()] = container[i]
                self.visit(ctx.children[4])
        except BreakIndicator:
            pass
        self.in_loop = False

    def visitNonPythonForStatement(self, ctx:PytovParser.nonPythonForStatement):
        if len(ctx.children) == 9:
            ctx.children = ctx.children[2:]
            ctx.children.pop(-2)

        self.visit(ctx.children[0])
        while_cond = ctx.children[2]
        self.in_loop = True
        try:
            while self.visit(while_cond):
                self.visit(ctx.children[-1]) # Block
                self.visit(ctx.children[-2]) # Execute at end of every iteration
        except BreakIndicator:
            pass
        self.in_loop = False


    # def visitOperator(self, ctx:PytovParser.OperatorContext):
    #     return ctx.getText()

    # def visitComparator(self, ctx:PytovParser.ComparatorContext):
    #     return ctx.getText()

    # def visitBinary(self, ctx:PytovParser.BinaryContext):
    #     return ctx.getText()

    def visitNotExpression(self, ctx:PytovParser.NotExpressionContext):
        return not self.visit(ctx.children[1])

    def visitParenExpression(self, ctx:PytovParser.ParenExpressionContext):
        return self.visit(ctx.exp())

    def visitPowExpression(self, ctx:PytovParser.PowExpressionContext):
        return self.visit(ctx.left) ** self.visit(ctx.right)

    def visitMultExpression(self, ctx:PytovParser.MultExpressionContext):
        return self.visit(ctx.left) * self.visit(ctx.right)

    def visitDivExpression(self, ctx:PytovParser.DivExpressionContext):
        return self.visit(ctx.left) / self.visit(ctx.right)

    def visitFdivExpression(self, ctx:PytovParser.FdivExpressionContext):
        return self.visit(ctx.left) // self.visit(ctx.right)
    
    def visitModExpression(self, ctx:PytovParser.ModExpressionContext):
        return self.visit(ctx.left) % self.visit(ctx.right)

    def visitAtom(self, ctx:PytovParser.AtomContext):
        return self.visit(ctx.children[0])

    def visitIncDec(self, ctx:PytovParser.IncDecContext):
        if type(ctx.children[1]) == PytovParser.IncContext:
            add = 1
        else:
            add = -1

        value = ctx.children[0].getText()
        if self.is_in_func:
            if value in list(self.active_funcs[-1].localVariables.keys()):
                self.active_funcs[-1].localVariables[value] += add
        elif value in list(self.variables.keys()):
            self.variables[value] += add
        elif value in list(self.global_variables.keys()):
            self.global_variables[value] += add
        elif value not in self.builtin_funcs:
            self.errorThrower._raise([ctx.start.line, ctx.getText()], "NameError", f"Name '{value}' is not defiend.")
        
        return self.visit(ctx.children[0]) - add

    def visitBeforeIncDec(self, ctx:PytovParser.BeforeIncDecContext):
        if type(ctx.children[0]) == PytovParser.IncContext:
            add = 1
        else:
            add = -1

        value = ctx.children[1].getText()
        if self.is_in_func:
            if value in list(self.active_funcs[-1].localVariables.keys()):
                self.active_funcs[-1].localVariables[value] += add
        elif value in list(self.variables.keys()):
            self.variables[value] += add
        elif value in list(self.global_variables.keys()):
            self.global_variables[value] += add
        elif value not in self.builtin_funcs:
            self.errorThrower._raise([ctx.start.line, ctx.getText()], "NameError", f"Name '{value}' is not defiend.")
        
        return self.visit(ctx.children[1])


    def visitOpCpBnExpression(self, ctx:PytovParser.OpCpBnExpressionContext):
        r = self.visit(ctx.right)
        l = self.visit(ctx.left)
        o = ctx.op.getText()

        if type(r) == str:
            r = '"' + r + '"'
        if type(l) == str:
            l = '"' + l + '"'

        if o == "+":
            return l + r
        elif o == "-":
            return l - r
        elif o == "*":
            return l * r
        elif o == "/":
            return l / r
        elif o == "/_":
            return l // r
        elif o == "%":
            return l % r
        elif o == ">":
            return l > r
        elif o == "<":
            return l < r
        elif o == ">=":
            return l >= r
        elif o == "<=":
            return l <= r
        elif o == "==":
            return l == r
        elif o == "!=":
            return l != r
        elif o == "&&":
            return l and r
        elif o == "||":
            return l or r
        else:
            return eval(' '.join([l, o, r]))

    def visitBoolTF(self, ctx:PytovParser.BoolTFContext):
        return ctx.getText().lower()[0] == "t"

    def visitIdentifier(self, ctx:PytovParser.IdentifierContext):
        value = ctx.getText()
        if self.is_in_func:
            if value in list(self.active_funcs[-1].localVariables.keys()):
                return self.active_funcs[-1].localVariables[value]
        elif value in list(self.variables.keys()):
            return self.variables[value]
        elif value in list(self.global_variables.keys()):
            return self.global_variables[value]
        elif value == "__name__":
            return self._name
        elif value not in self.builtin_funcs:
            self.errorThrower._raise([ctx.start.line, ctx.getText()], "NameError", f"Name '{value}' is not defiend.")
        
        self.errorThrower._raise([ctx.start.line, ctx.getText()], "NameError", f"Name '{value}' is not defiend.")

    def visitDecimal(self, ctx:PytovParser.DecimalContext):
        value = ctx.getText()
        if "." in value:
            return float(value)
        return int(value)

    def visitString(self, ctx:PytovParser.StringContext):
        return ctx.getText()[1:-1]

    def visitListr(self, ctx:PytovParser.ListrContext):
        return [self.visit(exp) for exp in ctx.children[1].children if type(exp) != TerminalNodeImpl]

    def visitDictr(self, ctx:PytovParser.DictrContext):
        return {self.visit(kv.children[0]):self.visit(kv.children[2]) for kv in ctx.children[1].children if type(kv) != TerminalNodeImpl}

    def visitVariableDeclaration(self, ctx:PytovParser.VariableDeclarationContext):
        if len(ctx.children) == 5:
            try:
                self.global_variables[ctx.children[1].getText()] = self.visit(ctx.children[3])
            except KeyError:
                self.errorThrower._raise([ctx.start.line, ctx.getText()], "NameError", f"(Global) Name '{ctx.children[1].getText()}' is not defiend.")

        if len(ctx.children) == 4:
            if type(ctx.children[0]) == PytovParser.IdentifierContext:
                op = ctx.children[1].getText()
                try:
                    # evaluate is unsafe, not too bad here because parser identified this as an operator, yet need to change in the future.
                    if ctx.children[0].getText() in self.global_variables:
                        self.global_variables[ctx.children[0].getText()] = eval(f"{self.global_variables[ctx.children[0].getText()]} {op} {self.visit(ctx.children[3])}")
                    elif self.is_in_func:
                        self.active_funcs[-1].localVariables[ctx.children[0].getText()] = eval(f"{self.active_funcs[-1].localVariables[ctx.children[0].getText()]} {op} {self.visit(ctx.children[3])}")
                    else:
                        self.variables[ctx.children[0].getText()] = eval(f"{self.variables[ctx.children[0].getText()]} {op} {self.visit(ctx.children[3])}")
                except KeyError:
                    self.errorThrower._raise([ctx.start.line, ctx.getText()], "NameError", f"Name '{ctx.children[0].getText()}' is not defiend.")
            else:
                self.global_variables[ctx.children[1].getText()] = self.visit(ctx.children[3])

        elif self.block_balance == 0 or self.block_is_true:
            if ctx.children[0].getText() in self.global_variables:
                self.global_variables[ctx.children[0].getText()] = self.visit(ctx.children[2])
            elif self.is_in_func:
                self.active_funcs[-1].localVariables[ctx.children[0].getText()] = self.visit(ctx.children[2])
            else:
                self.variables[ctx.children[0].getText()] = self.visit(ctx.children[2])

    def visitBlock(self, ctx:PytovParser.BlockContext):
        if (self.block_balance == 0) or (self.block_is_true):
            for blockItemList in ctx.children[1:-2]:
                self.visit(blockItemList)

    def visitOpenCurly(self, ctx):
        self.block_balance += 1
        self.last_block_is_true = self.block_is_true

    def visitCloseCurly(self, ctx):
        self.block_balance -= 1
        self.block_is_true = self.last_block_is_true


from antlr4 import *
from pytov.PytovLexer import PytovLexer
from pytov.PytovParser import PytovParser
from pytov.PytovVisitor import PytovVisitor
from antlr4.tree.Tree import TerminalNodeImpl
import argparse


def main(fileName, name='main'):
    input = FileStream(fileName)
    lexer = PytovLexer(input)
    stream = CommonTokenStream(lexer)
    parser = PytovParser(stream)
    tree = parser.parse()
    from pytov.PytovInterpreter import PytovInterpreter
    visitor = PytovInterpreter(fileName, name)
    visitor.visit(tree)
    return visitor

def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('file', help="File to run.")
    args = parser.parse_args()
    main(args.file)

if __name__ == '__main__':
    run()


if 1==2{
    x = 5-3
    if (1==1){
        y = "nooooo"
    }
    z = 8-x
}

// elif 1==1{
//     print("hihihi")
// }

if 1==2{
    print("not cool")
}
else if 1==3{
    print("very cool")
}
else{
    print("cooool")
}


function x(y,z, w=15){
    test = y+z+w
    print(test)
    return test
}

def func_test(){
    x(1, 2)
    print("fixed")
}

a = [1, 2, 3]
for i in range(1, 5){
    print(i)
    if i == 3{
        break
    }
}

z = x
print(z(1, 5, w=20))
func_test()

switch a{
    case [1, 2, 3]:
        print("hi0") // no break so will move on to case [1, 2, 3, 4] even though it is not equal to it.
    case [1, 2, 3, 4]:
        print("hi1")
        break
    default:
        print("hi2")
        break
    
}

#test

for (i = 5; i<10; i++){
    print(i)
}

x = 6
import "test2.txt"
print(x)


class ReturnIndicator(Exception):
    pass

class BreakIndicator(Exception):
    pass

class LineException:
    def __init__(self, file_name):
        self.file_name = file_name

    def _raise(self, line, err_type, massage):
        print(f"File '{self.file_name}' in line {line[0]}:\n  {line[1]}\n{err_type}: {massage}")
        exit()


from pytov.pytov import main, run
import pytov.PytovInterpreter
import pytov.PytovLexer
import pytov.PytovParser
import pytov.PytovListener
import pytov.PytovVisitor
import pytov.Objects
import pytov.exceptions


# Generated from Pytov.g4 by ANTLR 4.9.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .PytovParser import PytovParser
else:
    from PytovParser import PytovParser

# This class defines a complete generic visitor for a parse tree produced by PytovParser.

class PytovVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by PytovParser#parse.
    def visitParse(self, ctx:PytovParser.ParseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#allCode.
    def visitAllCode(self, ctx:PytovParser.AllCodeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#seperators.
    def visitSeperators(self, ctx:PytovParser.SeperatorsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#fdivExpression.
    def visitFdivExpression(self, ctx:PytovParser.FdivExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#powExpression.
    def visitPowExpression(self, ctx:PytovParser.PowExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#modExpression.
    def visitModExpression(self, ctx:PytovParser.ModExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#opCpBnExpression.
    def visitOpCpBnExpression(self, ctx:PytovParser.OpCpBnExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#notExpression.
    def visitNotExpression(self, ctx:PytovParser.NotExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#divExpression.
    def visitDivExpression(self, ctx:PytovParser.DivExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#parenExpression.
    def visitParenExpression(self, ctx:PytovParser.ParenExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#funcCallExpression.
    def visitFuncCallExpression(self, ctx:PytovParser.FuncCallExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#beforeIncDecExpression.
    def visitBeforeIncDecExpression(self, ctx:PytovParser.BeforeIncDecExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#atomExpression.
    def visitAtomExpression(self, ctx:PytovParser.AtomExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#multExpression.
    def visitMultExpression(self, ctx:PytovParser.MultExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#indexExpression.
    def visitIndexExpression(self, ctx:PytovParser.IndexExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#incDecExpression.
    def visitIncDecExpression(self, ctx:PytovParser.IncDecExpressionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#atom.
    def visitAtom(self, ctx:PytovParser.AtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#importp.
    def visitImportp(self, ctx:PytovParser.ImportpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#listr.
    def visitListr(self, ctx:PytovParser.ListrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#dictKV.
    def visitDictKV(self, ctx:PytovParser.DictKVContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#dictKVs.
    def visitDictKVs(self, ctx:PytovParser.DictKVsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#dictr.
    def visitDictr(self, ctx:PytovParser.DictrContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#incDec.
    def visitIncDec(self, ctx:PytovParser.IncDecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#beforeIncDec.
    def visitBeforeIncDec(self, ctx:PytovParser.BeforeIncDecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#inc.
    def visitInc(self, ctx:PytovParser.IncContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#dec.
    def visitDec(self, ctx:PytovParser.DecContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#opCpBn.
    def visitOpCpBn(self, ctx:PytovParser.OpCpBnContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#notl.
    def visitNotl(self, ctx:PytovParser.NotlContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#identifier.
    def visitIdentifier(self, ctx:PytovParser.IdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#decimal.
    def visitDecimal(self, ctx:PytovParser.DecimalContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#string.
    def visitString(self, ctx:PytovParser.StringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#comparator.
    def visitComparator(self, ctx:PytovParser.ComparatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#binary.
    def visitBinary(self, ctx:PytovParser.BinaryContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#operator.
    def visitOperator(self, ctx:PytovParser.OperatorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#index.
    def visitIndex(self, ctx:PytovParser.IndexContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#indexOverwrite.
    def visitIndexOverwrite(self, ctx:PytovParser.IndexOverwriteContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#boolTF.
    def visitBoolTF(self, ctx:PytovParser.BoolTFContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#ifStatement.
    def visitIfStatement(self, ctx:PytovParser.IfStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#breakp.
    def visitBreakp(self, ctx:PytovParser.BreakpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#switchCase.
    def visitSwitchCase(self, ctx:PytovParser.SwitchCaseContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#switchDefault.
    def visitSwitchDefault(self, ctx:PytovParser.SwitchDefaultContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#switchStatement.
    def visitSwitchStatement(self, ctx:PytovParser.SwitchStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#identifierList.
    def visitIdentifierList(self, ctx:PytovParser.IdentifierListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#declaration.
    def visitDeclaration(self, ctx:PytovParser.DeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#declarationList.
    def visitDeclarationList(self, ctx:PytovParser.DeclarationListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#variableDeclaration.
    def visitVariableDeclaration(self, ctx:PytovParser.VariableDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#expList.
    def visitExpList(self, ctx:PytovParser.ExpListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#parameterList.
    def visitParameterList(self, ctx:PytovParser.ParameterListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#parameterInputList.
    def visitParameterInputList(self, ctx:PytovParser.ParameterInputListContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#whileStatement.
    def visitWhileStatement(self, ctx:PytovParser.WhileStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#forStatement.
    def visitForStatement(self, ctx:PytovParser.ForStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#nonPythonForStatement.
    def visitNonPythonForStatement(self, ctx:PytovParser.NonPythonForStatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#funcDeclaration.
    def visitFuncDeclaration(self, ctx:PytovParser.FuncDeclarationContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#funcCall.
    def visitFuncCall(self, ctx:PytovParser.FuncCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#methodCall.
    def visitMethodCall(self, ctx:PytovParser.MethodCallContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#returnp.
    def visitReturnp(self, ctx:PytovParser.ReturnpContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#statement.
    def visitStatement(self, ctx:PytovParser.StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#block.
    def visitBlock(self, ctx:PytovParser.BlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#openCurly.
    def visitOpenCurly(self, ctx:PytovParser.OpenCurlyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PytovParser#closeCurly.
    def visitCloseCurly(self, ctx:PytovParser.CloseCurlyContext):
        return self.visitChildren(ctx)



del PytovParser